"""
Wrapper code to make it easy to run experiments with LAAOS.

`parse_args_and_load_experiment_config` facilitates loading external experiment configs.

`embedded_experiment` returns a store for an experiment based on the current script file.

`embedded_experiments` returns a store a given job (out of number of jobs).
"""
import argparse
import functools
import os
import sys
import time
from os import path

import dataclasses

import torch
import typing
from laaos import TypeHandler

import laaos
import laaos.torch

# KEEP THIS AROUND TO SET THE DIRECTORY AUTOMAGICALLY
import blackhc.project.script


def get_likely_github_commit_url(github_url: str, commit):
    # "git@github.com:user/repo.git"
    # "https://github.com/user/repo.git"

    git_prefix = "git@github.com:"
    if github_url.startswith(git_prefix):
        github_url = f"https://github.com/{github_url[len(git_prefix)]:}"
    https_prefix = "https://github.com/"
    if not github_url.startswith(https_prefix):
        return commit

    suffix = ".git"
    if github_url.endswith(suffix):
        github_url = f"{github_url[:-len(suffix)]}"
    else:
        return commit

    return f"{github_url}/commit/{commit}"


def get_git_head_commit_and_url(path):
    try:
        import git
    except ImportError:
        return "", ""

    try:
        repo = git.Repo(path, search_parent_directories=True)
        commit = str(repo.commit())
        urls = [get_likely_github_commit_url(url, commit) for url in repo.remote("origin").urls] or [commit]
        return commit, urls[0]
    except git.exc.InvalidGitRepositoryError:
        return "", ""


def dummy_add_experiment_config_args(parser):
    return parser


def parse_args_and_load_experiment_config(
    add_experiment_config_args=dummy_add_experiment_config_args, exposed_symbols=tuple(), description="Experiment"
):
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=functools.partial(argparse.ArgumentDefaultsHelpFormatter, width=120),
    )
    parser.add_argument("--experiment_id", type=str, default=None, help="experiment id")
    parser.add_argument(
        "--experiment_config", type=str, default=None, help="Laaos file that contains all experiment task configs"
    )
    parser.add_argument(
        "--result_dir", type=str, default=None, help="Laaos file that contains all experiment task configs"
    )
    parser.add_argument(
        "--experiment_description", type=str, default="Trying stuff..", help="Description of the experiment"
    )
    parser = add_experiment_config_args(parser)
    args = parser.parse_args()

    if args.experiments_config is not None:
        config = laaos.safe_load(args.experiments_config, exposed_symbols=exposed_symbols)
        # Merge the experiment config with args.
        # Args take priority.
        args = parser.parse_args(namespace=argparse.Namespace(**config[args.experiment_id]))

    # DONT TRUNCATE LOG FILES EVER AGAIN!!! (OFC THIS HAD TO HAPPEN AND WAS PAINFUL)
    if args.experiment_id:
        store_name = f"experiment_{args.experiment_id}_result"
    else:
        store_name = f"unnamed_result"

    if args.experiments_config:
        result_dir = path.dirname(args.experiments_config) + "/results/"
    else:
        result_dir = "./experiments/runs/results"

    store = create_experiment_store(result_dir=result_dir, store_name=store_name)
    store["args"] = args.__dict__
    store["experiments_config"] = args.experiments_config
    store["experiment_id"] = args.experiment_id

    print("Parsed args:")
    print(args.__dict__)

    return args, store


def embedded_experiment(script_file):
    script_file = path.abspath(script_file)
    result_dir = path.dirname(script_file) + "/results/"
    result_name = path.splitext(path.basename(script_file))[0] + "_result"
    store = create_experiment_store(result_dir=result_dir, store_name=result_name)
    store["experiment"] = script_file
    return store


def embedded_experiments(script_file, num_jobs):
    parser = argparse.ArgumentParser(
        description="Unpacking Information Bottlenecks Experiment",
        formatter_class=functools.partial(argparse.ArgumentDefaultsHelpFormatter, width=120),
    )
    parser.add_argument("--id", type=int, default=None, help="experiment id")
    parser.add_argument(
        "--num_workers", type=int, default=None, help="number of worker (None means one worker per job)"
    )
    args = parser.parse_args()

    num_workers = args.num_workers or num_jobs
    worker_id = args.id
    job_id = args.id
    if job_id is None or not (0 <= job_id < num_jobs):
        raise ValueError(f"0 <= --id={job_id} < {num_jobs}!")

    # DONT TRUNCATE LOG FILES EVER AGAIN!!! (OFC THIS HAD TO HAPPEN AND BE PAINFUL)
    script_file = path.abspath(script_file)
    result_dir = path.dirname(script_file) + "/results/"

    while job_id < num_jobs:
        result_name = path.splitext(path.basename(script_file))[0] + f"_job_{job_id}"
        store = create_experiment_store(result_dir=result_dir, store_name=result_name)

        store["experiment"] = script_file
        store["job_id"] = job_id
        store["worker_id"] = worker_id
        store["num_workers"] = num_workers

        print(f"Job id: {job_id}/{num_jobs}")
        print(f"Worker id: {worker_id}/{num_workers}")

        yield job_id, store

        store.close()

        job_id += num_workers


def create_experiment_store(*, result_dir, store_name=None):
    # Make sure we have a directory to store the results in, and we don't crash!
    os.makedirs(result_dir, exist_ok=True)
    safe_store_name = store_name if store_name else "results"
    store = laaos.open_file_store(
        safe_store_name,
        prefix=result_dir,
        truncate=False,
        type_handlers=laaos.torch.TypeHandlers + laaos.DefaultTypeHandlers,
    )

    store["timestamp"] = int(time.time())
    store["cmdline"] = sys.argv[:]
    commit, github_url = get_git_head_commit_and_url(".")
    store["commit"] = commit
    store["github_url"] = github_url
    print("Command line:")
    print("|".join(sys.argv))
    print(f"GitHub URL: {github_url}")
    print(f"Commit: {commit}")
    print(f"Results stored in {store.uri}")
    return store



import sys
import os
import cgitb

project_dir = None
original_dir = None


def set_project_dir(new_project_dir):
    global original_dir
    global project_dir

    if not original_dir:
        original_dir = os.getcwd()

    if project_dir:
        old_src_path = get_src_path(project_dir)
        if old_src_path in sys.path:
            sys.path.remove(old_src_path)

    project_dir = new_project_dir

    if not project_dir:
        print('Neither src found as subdirectory in %s nor was a notebooks directory found!' % original_dir)
        return

    src_path = get_src_path(project_dir)
    if src_path not in sys.path:
        sys.path.append(src_path)
        print('Appended %s to paths' % src_path)

    switch_to_project_dir()


def switch_to_project_dir():
    os.chdir(project_dir)
    print('Switched to directory %s' % project_dir)


def get_src_path(project_dir):
    return os.path.join(project_dir, 'src')


# Inspired by https://stackoverflow.com/questions/19687394/python-script-to-determine-if-a-directory-is-a-git-repository
def get_git_working_dir(path):
    try:
        import git
    except ImportError:
        return None

    try:
        return git.Repo(path, search_parent_directories=True).working_tree_dir
    except git.exc.InvalidGitRepositoryError:
        return None


def get_cookiecutter_project_path(seed_path):
    # Check if we are within a notebooks sub path
    path = seed_path
    while not os.path.ismount(path):
        if os.path.basename(path) in ['notebooks', 'scripts']:
            project_dir = os.path.dirname(path)
            return project_dir
        path = os.path.dirname(path)

    # Check if we are at the root. And there is a src subdirectory.
    if os.path.isdir(os.path.join(seed_path, 'src')):
        return os.path.abspath(seed_path)

    return None


def infer_and_set_project_dir():
    set_project_dir(get_git_working_dir(os.getcwd()) or get_cookiecutter_project_path(os.getcwd()))


def echo_magic(magic):
    from IPython import get_ipython
    ipython = get_ipython()
    print('%%%s' % magic)
    ipython.magic(magic)


def setup_autoreload():
    echo_magic('load_ext autoreload')
    echo_magic('autoreload 2')


_is_run_from_ipython = None


def is_run_from_ipython():
    global _is_run_from_ipython

    if _is_run_from_ipython is not None:
        return _is_run_from_ipython

    try:
        # noinspection PyUnresolvedReferences
        __IPYTHON__
        _is_run_from_ipython = True
    except NameError:
        _is_run_from_ipython = False

    return _is_run_from_ipython


def enable_detailed_tracebacks():
    cgitb.enable(format="text")



from blackhc.project import infer_and_set_project_dir, is_run_from_ipython, setup_autoreload

infer_and_set_project_dir()

if is_run_from_ipython():
    # Only execute auto_reload if we are in ipython
    setup_autoreload()


# Always prefer setuptools over distutils
from setuptools import setup
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='blackhc.project',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='2.5.1',

    description='Notebook setup code',
    # Fix windows newlines.
    long_description=long_description.replace('\r\n', '\n'),
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url='https://github.com/blackhc/notebook_setup',

    # Author details
    author='Andreas @blackhc Kirsch',
    author_email='blackhc+notebook_setup@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    # What does your project relate to?
    keywords='jupyter',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=['blackhc.project', 'blackhc.project.utils'],

    package_dir={'': 'src'},

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['gitpython', 'fs.sshfs', 'laaos'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage', 'pytest', 'pytest-forked', 'pyfakefs==3.3', 'torch', 'psutil', 'pytorch-ignite',
                 'blackhc.progress_bar'],
    },

    setup_requires=[
        'pytest-runner',
    ],
)



import ignite
from blackhc.progress_bar import create_progress_bar


class IgniteProgressBar(object):
    def __init__(self, desc, log_interval):
        self.log_interval = log_interval
        self.desc = desc
        self.progress_bar = None

    def attach(self, engine: ignite.engine.Engine):
        engine.add_event_handler(ignite.engine.Events.EPOCH_STARTED, self.on_start)
        engine.add_event_handler(ignite.engine.Events.EPOCH_COMPLETED, self.on_complete)
        engine.add_event_handler(ignite.engine.Events.ITERATION_COMPLETED, self.on_iteration_complete)

    def on_start(self, engine):
        dataloader = engine.state.dataloader
        self.progress_bar = create_progress_bar(len(dataloader))

        print(self.desc(engine) if callable(self.desc) else self.dec)
        self.progress_bar.start()

    def on_complete(self, engine):
        self.progress_bar.finish()

    def on_iteration_complete(self, engine):
        dataloader = engine.state.dataloader
        iter = (engine.state.iteration - 1) % len(dataloader) + 1

        if iter % self.log_interval == 0:
            self.progress_bar.update(self.log_interval)


def ignite_progress_bar(engine: ignite.engine.Engine, desc=None, log_interval=0):
    wrapper = IgniteProgressBar(desc, log_interval)
    wrapper.attach(engine)

    return wrapper
